@@ -24,6 +24,8 @@
 
 #include <QtGui>
 
+class KeyMenu;
+
 //fix tearing during scrolling
 class ScrollArea : public QScrollArea {
 public:
@@ -37,6 +39,7 @@ protected:
 	}
 };
 
+
 class MainWindow : public QMainWindow {
 	Q_OBJECT
 public:
@@ -44,7 +47,7 @@ public:
 public slots:
 	void about();
 	void showConnectDialog();
-	void connectToHost(QString url);
+	void connectToHost(QString url, int quality);
 	void disconnectFromHost();
 	void forceResize();
 	void forceResizeDelayed();
@@ -54,12 +57,13 @@ public slots:
 	void sendPgDn() { vnc_view->sendKey(Qt::Key_PageDown); }
 	void sendReturn() { vnc_view->sendKey(Qt::Key_Return); }
 	void showInputPanel();
-	void showModifierMenu();
+	void showKeyMenu();
 	void showPreferences();
 	void statusChanged(RemoteView::RemoteStatus status);
 	void toggleFullscreen();
 protected:
 	void closeEvent(QCloseEvent*);
+	void resizeEvent(QResizeEvent *event);
 private:
 	void grabZoomKeys(bool grab);
 	void reloadSettings();
@@ -67,5 +71,6 @@ private:
 	ScrollArea *scroll_area;
 	QToolBar *toolbar;	
 	QAction *scaling, *show_toolbar, *disconnect_action;
+	KeyMenu *key_menu;
 };
 #endif
